//  -*- C++ -*-
//
// Methods for Markov Decision Problems
//
// The MDP class defines a set of methods for manipulating Markov
// Decision Problems including value iteration, policy iteration, and
// prioritized sweeping using tabular methods.
//
// All of these classes are parameterized using templates in terms of
// two types:  State and Action.  Values of these types are passed to
// methods of the Problem class in order to manipulate them.

#ifndef MDP_H
#define MDP_H

#include "vector.h"
#include "list.h"

extern int VERBOSE;
extern int BackupCounter;  // count number of "Q" backups.


template <class State>
class ResultInfo
{
public:
  ResultInfo() {};
  ResultInfo(float p, float r, State s): probability(p),
					 reward(r),
					 state(s) {};
  float probability;
  float reward;
  State state;
};

template <class State>
istream & operator >> (istream &, ResultInfo<State> &);
template <class State>
ostream & operator << (ostream &, ResultInfo<State> &);


template <class State, class Action>
class SuccessorInfo
{
public:
  // constructors
  SuccessorInfo() {};
  SuccessorInfo(Action a): action(a) {};
  Action action;
  list<ResultInfo<State> *> resultStates;
};

template <class State, class Action>
istream & operator >> (istream &, SuccessorInfo<State, Action> &);
template <class State, class Action>
ostream & operator << (ostream &, SuccessorInfo<State, Action> &);
  
template <class State, class Action>
class PredecessorInfo
{
public:
  PredecessorInfo(State s, Action a): state(s), action(a) {};
  State  state;
  Action  action;
};

template <class State, class Action>
class Problem
{
public:
  virtual float discountFactor() = 0;
  virtual unsigned int maxState() = 0;
  virtual unsigned int maxAction() = 0;
  virtual iterator<State> * stateIterator() = 0;
  virtual list<SuccessorInfo<State, Action> *> * successors(State) = 0;
  virtual list<PredecessorInfo<State, Action> *> * predecessors(State) = 0;
  virtual unsigned int stateIndex(State) = 0;
  virtual unsigned int actionIndex(Action) = 0;
  virtual void printState(ostream & str, State & s) = 0;
  virtual void printAction(ostream & str, Action & a) = 0;
};

template <class State, class Action>
class MDP
{
public:
  MDP(Problem<State, Action> * p);

  float Vfunction(State);
  float & Qfunction(State, Action);

  void Qbackup(State, Action);
  float BellmanBackup(State &);
  void ValueIteration(float epsilon);
  void PrintValueFunction(ostream &);
  Action greedyPolicy(State);

protected:
  vector<float> Q;                  // the Q table
  Problem<State, Action> * problem; // the problem we are solving
  unsigned int nActions;         // the max num of actions, for fast indexing
};

// implementations

template <class State, class Action>
MDP<State, Action>::MDP(Problem<State, Action> * p):
  problem(p),
  Q(p->maxState() * p->maxAction(), 0.0),
  nActions(p->maxAction())
{
}
  
template <class State, class Action>
float MDP<State, Action>::Vfunction(State state)
{
  list<SuccessorInfo<State, Action> *> * succs = problem->successors(state);
  assert(succs);
  listIterator<SuccessorInfo<State, Action> *> itr(*succs);
  float answer = 0.0;
  int firstTime = 1;
  unsigned int base = nActions * problem->stateIndex(state);
  for (itr.init(); !itr; ++itr) {
    unsigned int index = problem->actionIndex(itr()->action) + base;
    float qvalue = Q[index];
    if (firstTime) {
      answer = qvalue;
      firstTime = 0;
    }
    else if (qvalue > answer) answer = qvalue;
  }
  return answer;
}

template <class State, class Action>
float & MDP<State, Action>::Qfunction(State state, Action action)
{
  unsigned int index = (problem->stateIndex(state) * nActions +
			problem->actionIndex(action));
  return Q[index];
}

template <class State, class Action>
void MDP<State, Action>::Qbackup(State state, Action action)
{
  list<SuccessorInfo<State, Action> *> * succs = problem->successors(state);
  assert(succs);

  // find the desired action in the successor information
  listIterator<SuccessorInfo<State, Action> *> itr(*succs);
  for (itr.init(); !itr; ++itr) {
    if (itr()->action == action) {
      float answer = 0.0;
      
      // loop through successor states and compute Bellman formula
      listIterator<ResultInfo<State> *> ritr(itr()->resultStates);
      for(ritr.init(); !ritr; ++ritr) {
	answer += ritr()->probability * (ritr()->reward +
					 problem->discountFactor() *
					 Vfunction(ritr()->state));
      }
      Qfunction(state, action) = answer;
      BackupCounter++;
      return;
    }
  }
}

//
// Perform a bellman backup
// Returns the bellman error
//
template <class State, class Action>
float MDP<State, Action>::BellmanBackup(State & state)
{
  if (VERBOSE) {
    cout << "Starting BellmanBackup for state ";
    problem->printState(cout, state);
    cout << endl;
  }

  // we update the Q value for each action
  list<SuccessorInfo<State, Action> *> * succs = problem->successors(state);
  assert(succs);
  float oldValue = 0.0;
  float newValue = 0.0;
  int firstTime = 1;

  listIterator<SuccessorInfo<State, Action> *> itr(*succs);
  for (itr.init(); !itr; ++itr) {
    if (VERBOSE) cout << "  Action = " << itr()->action << endl;
    float qvalue = 0.0;
      
    // loop through successor states and compute Bellman formula
    listIterator<ResultInfo<State> *> ritr(itr()->resultStates);
    for(ritr.init(); !ritr; ++ritr) {
      if (VERBOSE) {
	cout << "    prob=" << ritr()->probability
	     << " reward=" << ritr()->reward
	     << " state'=" << ritr()->state
	     << " V(state') = " << Vfunction(ritr()->state)
	     << endl;
      }
      qvalue += ritr()->probability * (ritr()->reward +
				       problem->discountFactor() *
				       Vfunction(ritr()->state));
    }

    // compute new and old values
    float & cell = Qfunction(state, itr()->action);
    if (VERBOSE) {
      cout << "  Qold = " << cell << " Qnew = " << qvalue << endl;
    }

    if (firstTime) {
      oldValue = cell;
      newValue = qvalue;
      firstTime = 0;
    }
    else {
      if (cell > oldValue) oldValue = cell;
      if (qvalue > newValue) newValue = qvalue;
    }
    // this updates the Qvalue
    cell = qvalue;
    BackupCounter++;
  }
  // note: if there were no actions at all, this will return 0.0 for
  // the Bellman error.
  if (VERBOSE) {
    cout << "Bellman error=" << fabs(oldValue - newValue) << endl;
  }
  return fabs(oldValue - newValue);
}
  
template <class State, class Action>
void MDP<State, Action>::ValueIteration(float epsilon)
{
  iterator<State> & itr = *(problem->stateIterator());
  
  int epsilonExceeded = 0;
  int iteration = 0;
  do {
    iteration++;
    epsilonExceeded = 0;
    float totalBellmanError = 0.0;
    for (itr.init(); !itr; ++itr) {
      float bellmanerror = BellmanBackup(itr());
      if (bellmanerror > epsilon) epsilonExceeded = 1;
      totalBellmanError += bellmanerror;
    }
    cout << "End of iteration " << iteration
	 << " BackupCounter = " << BackupCounter
	 << " total bellman error = " << totalBellmanError << endl;
    if (VERBOSE) PrintValueFunction(cout);
  }
  while (epsilonExceeded);
}


// IO routines
template <class State>
istream & operator >> (istream & str, ResultInfo<State> & ri) 
{
  char delim;
  str >> delim;
  if (delim != '(') {
    cerr << "While reading ResultInfo, expected '(' but read '" << delim << "'." << endl;
    abort();
  }
  str >> ri.probability;
  str >> ri.reward;
  str >> ri.state;
  str >> delim;
  if (delim != ')') {
    cerr << "while reading ResultInfo, expected ')' but read '" <<
      delim << "'." << endl;
    abort();
  }
  return str;
}

template <class State>
istream & operator >> (istream & str, ResultInfo<State> * ri) 
{
  return str >> (*ri);
}

template <class State>
ostream & operator << (ostream & str, ResultInfo<State> & ri)
{
  str << "(" << ri.probability << " "
      << ri.reward << " "
      << ri.state << ")";
  return str;
}

template <class State>
ostream & operator << (ostream & str, ResultInfo<State> * ri)
{
  return str << *ri;
}

template <class State, class Action>
istream & operator >> (istream & str, SuccessorInfo<State, Action> & si)
{
  char delim;
  str >> delim;
  if (delim != '(') {
    cerr << "While reading SuccessorInfo expected '(' but read '"
	 << delim << "'." << endl;
    abort();
  }
  str >> si.action;
  str >> si.resultStates;
  str >> delim;
  if (delim != ')') {
    cerr << "While reading SuccessorInfo expected ')' but read '"
	 << delim << "'." << endl;
  }
  return str;
}

template <class State, class Action>
istream & operator >> (istream & str, SuccessorInfo<State, Action> * si)
{
  return str >> (*si);
}

template <class State, class Action>
ostream & operator << (ostream & str, SuccessorInfo<State, Action> & si)
{
  str << "(" << si.action << " " << si.resultStates << ")";
  return str;
}

template <class State, class Action>
ostream & operator << (ostream & str, SuccessorInfo<State, Action> * 
		       si)
{
  return str << *si;
}

template <class State, class Action>
istream & operator >> (istream & istr,
		       list<SuccessorInfo<State, Action> *> & target)
{
  // read in a list of objects surrounded by parentheses and separated by spaces
  // we push them onto the target list and then reverse the list
  assert(target.isEmpty()); // target must be empty

  char delim;
  istr >> delim;
  if (delim != '(') {
    cerr << "Found `" << delim << "' where `(' was expected." << endl;
  }
  assert(delim == '(');
  do {
    istr >> delim;
    if (delim == ')') break;
    istr.putback(delim);
    SuccessorInfo<State, Action> * val = new SuccessorInfo<State, Action>;
    istr >> val;
    target.add(val);
  } while (istr);
  target.reverse();
  return istr;
}

template <class State>
istream & operator >> (istream & istr,
		       list<ResultInfo<State> *> & target)
{
  // read in a list of objects surrounded by parentheses and separated by spaces
  // we push them onto the target list and then reverse the list
  assert(target.isEmpty()); // target must be empty

  char delim;
  istr >> delim;
  if (delim != '(') {
    cerr << "Found `" << delim << "' where `(' was expected." << endl;
  }
  assert(delim == '(');
  do {
    istr >> delim;
    if (delim == ')') break;
    istr.putback(delim);
    ResultInfo<State> * val = new ResultInfo<State>;
    istr >> val;
    target.add(val);
  } while (istr);
  target.reverse();
  return istr;
}

template <class State, class Action>
void MDP<State,Action>::PrintValueFunction(ostream & str)
{
  iterator<State> & itr = *(problem->stateIterator());

  for(itr.init(); !itr; ++itr) {
    problem->printState(str, itr());
    cout << " " << Vfunction(itr()) << endl;
  }
}

template <class State, class Action>
Action MDP<State,Action>::greedyPolicy(State state)
{
  list<SuccessorInfo<State, Action> *> * succs = problem->successors(state);
  assert(succs);
  float bestValue = 0.0;
  Action bestAction;
  int firstTime = 1;
  listIterator<SuccessorInfo<State, Action> *> itr(*succs);
  for (itr.init(); !itr; ++itr) {
    float qvalue = Qfunction(state, itr()->action);
    if (firstTime) {
      bestValue = qvalue;
      bestAction = itr()->action;
      firstTime = 0;
    }
    else if (qvalue > bestValue) {
      bestValue = qvalue;
      bestAction = itr()->action;
    }
  }
  return bestAction;
}


  




#endif
